@@ -24,6 +24,7 @@
 // (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 // SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
+const Gc = imports.gi.Gc;
 const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
 const GObject = imports.gi.GObject;
@@ -39,12 +40,12 @@ const Gettext = imports.gettext;
 const Main = imports.main;
 const Util = imports.util;
 
-const MainWindow = new Lang.Class({
+var MainWindow = new Lang.Class({
     Name: 'MainWindow',
     Extends: Gtk.ApplicationWindow,
     Template: 'resource:///org/gnome/Characters/mainwindow.ui',
     InternalChildren: ['main-headerbar', 'search-active-button',
-                       'search-bar', 'search-entry',
+                       'search-bar', 'search-entry', 'back-button',
                        'menu-button',
                        'main-grid', 'main-hbox', 'sidebar-grid'],
     Properties: {
@@ -69,10 +70,16 @@ const MainWindow = new Lang.Class({
                             activate: this._toggleSearch,
                             parameter_type: new GLib.VariantType('b'),
                             state: new GLib.Variant('b', false) },
+                          { name: 'find',
+                            activate: this._find },
                           { name: 'category',
                             activate: this._category,
                             parameter_type: new GLib.VariantType('s'),
-                            state: new GLib.Variant('s', 'punctuation') },
+                            state: new GLib.Variant('s', 'emojis') },
+                          { name: 'subcategory',
+                            activate: this._subcategory,
+                            parameter_type: new GLib.VariantType('s'),
+                            state: new GLib.Variant('s', 'emoji-smileys') },
                           { name: 'character',
                             activate: this._character,
                             parameter_type: new GLib.VariantType('s') },
@@ -80,6 +87,8 @@ const MainWindow = new Lang.Class({
                             activate: this._filterFont,
                             parameter_type: new GLib.VariantType('s') }]);
 
+        this.application.set_accels_for_action('win.find', ['<Primary>F']);
+
         this.bind_property('search-active', this._search_active_button, 'active',
                            GObject.BindingFlags.SYNC_CREATE |
                            GObject.BindingFlags.BIDIRECTIONAL);
@@ -92,24 +101,31 @@ const MainWindow = new Lang.Class({
         this._search_entry.connect('search-changed',
                                    Lang.bind(this, this._handleSearchChanged));
 
+        this._back_button.connect('clicked',
+                                  Lang.bind(this, function() {
+                                      let action = this.lookup_action('category');
+                                      action.activate(new GLib.Variant('s', 'emojis'));
+                                  }));
+        this._back_button.bind_property('visible',
+                                        this._search_active_button, 'visible',
+                                        GObject.BindingFlags.SYNC_CREATE |
+                                        GObject.BindingFlags.INVERT_BOOLEAN);
+
         this._menu_popover = new Menu.MenuPopover({});
         this._menu_button.set_popover(this._menu_popover);
 
-        this._categoryList =
-            new CategoryList.CategoryListWidget({ vexpand: true });
+        this._categoryListView =
+            new CategoryList.CategoryListView({ vexpand: true });
         let scroll = new Gtk.ScrolledWindow({
             hscrollbar_policy: Gtk.PolicyType.NEVER,
             hexpand: false,
         });
-        scroll.add(this._categoryList);
+        scroll.add(this._categoryListView);
         this._sidebar_grid.add(scroll);
 
-        this._mainView = new MainView({ categoryList: this._categoryList });
-
-        if (this._mainView.recentCharacters.length == 0) {
-            let row = this._categoryList.get_row_at_index(1);
-            this._categoryList.select_row(row);
-        }
+        this._mainView = new MainView({
+            categoryListView: this._categoryListView
+        });
 
         this._main_hbox.pack_start(this._mainView, true, true, 0);
         this._main_grid.show_all();
@@ -119,6 +135,22 @@ const MainWindow = new Lang.Class({
         this.connect('key-press-event', Lang.bind(this, this._handleKeyPress));
     },
 
+    vfunc_map: function() {
+        this.parent();
+        this._selectFirstSubcategory();
+    },
+
+    // Select the first subcategory which contains at least one character.
+    _selectFirstSubcategory: function() {
+        let categoryList = this._categoryListView.get_visible_child();
+        let index = 0;
+        let row = categoryList.get_row_at_index(index);
+        if (row.category.name == 'recent' &&
+            this._mainView.recentCharacters.length == 0)
+            index++;
+        categoryList.select_row(categoryList.get_row_at_index(index));
+    },
+
     get search_active() {
         return this._searchActive;
     },
@@ -129,8 +161,10 @@ const MainWindow = new Lang.Class({
 
         this._searchActive = v;
 
-        if (this._searchActive)
-            this._categoryList.unselect_all();
+        if (this._searchActive) {
+            let categoryList = this._categoryListView.get_visible_child();
+            categoryList.unselect_all();
+        }
 
         this.notify('search-active');
     },
@@ -196,13 +230,41 @@ const MainWindow = new Lang.Class({
 
         let [name, length] = v.get_string()
 
-        let category = this._categoryList.getCategory(name);
+        this._categoryListView.set_visible_child_name(name);
+        let categoryList = this._categoryListView.get_visible_child();
+        if (categoryList == null)
+            return;
+
+        this._selectFirstSubcategory();
+        let category = categoryList.get_selected_row().category;
+
+        if (name == 'emojis') {
+            this._back_button.hide();
+        } else {
+            this._back_button.show();
+        }
 
         Util.assertNotEqual(category, null);
         this._mainView.setPage(category);
         this._updateTitle(category.title);
     },
 
+    _subcategory: function(action, v) {
+        this.search_active = false;
+
+        let [name, length] = v.get_string()
+
+        let categoryList = this._categoryListView.get_visible_child();
+        if (categoryList == null)
+            return;
+
+        let category = categoryList.getCategory(name);
+        if (category) {
+            this._mainView.setPage(category);
+            this._updateTitle(category.title);
+        }
+    },
+
     _character: function(action, v) {
         let [uc, length] = v.get_string()
         this._mainView.addToRecent(uc);
@@ -217,6 +279,10 @@ const MainWindow = new Lang.Class({
         this._menu_popover.hide();
     },
 
+    _find: function() {
+        this.search_active = !this.search_active;
+    },
+
     setSearchKeywords: function(keywords) {
         this.search_active = keywords.length > 0;
         this._search_entry.set_text(keywords.join(' '));
@@ -226,6 +292,7 @@ const MainWindow = new Lang.Class({
 const MainView = new Lang.Class({
     Name: 'MainView',
     Extends: Gtk.Stack,
+    Template: 'resource:///org/gnome/Characters/mainview.ui',
     Properties: {
         'max-recent-characters': GObject.ParamSpec.uint(
             'max-recent-characters', '', '',
@@ -250,31 +317,61 @@ const MainView = new Lang.Class({
 
     set filterFontFamily(family) {
         this._filterFontFamily = family;
-        this.visible_child.setFilterFont(family);
+        this._fontFilter.setFilterFont(this._filterFontFamily);
     },
 
     _init: function(params) {
-        let filtered = Params.filter(params, { categoryList: null });
+        let filtered = Params.filter(params, { categoryListView: null });
         params = Params.fill(params, {
             hexpand: true, vexpand: true,
             transition_type: Gtk.StackTransitionType.CROSSFADE
         });
         this.parent(params);
 
+        this._fontFilter = new CharacterList.FontFilter({});
         this._filterFontFamily = null;
         this._characterLists = {};
-        this._categoryList = filtered.categoryList;
+        this._recentCharacterLists = {};
+        this._categoryListView = filtered.categoryListView;
 
         let characterList;
-        let categories = this._categoryList.getCategoryList();
-        for (let index in categories) {
-            let category = categories[index];
-            characterList = this._createCharacterList(
-                category.name, _('%s Character List').format(category.title));
-            // FIXME: Can't use GtkContainer.child_get_property.
-            characterList.title = category.title;
-            this.add_titled(characterList, category.name, category.title);
+        let categories = this._categoryListView.getCategoryList();
+        let recentBox = new Gtk.Box({ orientation: Gtk.Orientation.VERTICAL,
+                                      hexpand: true, vexpand: false });
+
+        for (let i in categories) {
+            let category = categories[i];
+            let categoryList = this._categoryListView.get_child_by_name(category.name);
+            let subcategories = categoryList.getCategoryList();
+            for (let j in subcategories) {
+                let subcategory = subcategories[j];
+                characterList = this._createCharacterList(
+                    subcategory.name,
+                    _('%s Character List').format(subcategory.title));
+                // FIXME: Can't use GtkContainer.child_get_property.
+                characterList.title = subcategory.title;
+                this.add_titled(characterList, subcategory.name, subcategory.title);
+            }
+            characterList = this._createRecentCharacterList(
+                category.name,
+                // TRANSLATORS: %s will be either 'emojis' or 'letters'
+                _('Recently Used %s Character List').format(category.title),
+                category.category);
+            this._recentCharacterLists[category.name] = characterList;
+            if (i > 0) {
+                let separator = new Gtk.Separator({});
+                recentBox.pack_start(separator, false, false, 0);
+            }
+            recentBox.pack_start(characterList, false, false, 0);
         }
+        let scroll = new Gtk.ScrolledWindow({
+            hscrollbar_policy: Gtk.PolicyType.NEVER,
+            hexpand: false,
+        });
+        scroll.add(recentBox);
+        // FIXME: Can't use GtkContainer.child_get_property.
+        scroll.title = _('Recently Used');
+        this.add_titled(scroll, 'recent', scroll.title);
 
         characterList = this._createCharacterList(
             'search-result', _('Search Result Character List'));
@@ -292,13 +389,25 @@ const MainView = new Lang.Class({
     },
 
     _createCharacterList: function(name, accessible_name) {
-        let characterList = new CharacterList.CharacterListView({});
+        let characterList = new CharacterList.CharacterListView({
+            fontFilter: this._fontFilter
+        });
         characterList.get_accessible().accessible_name = accessible_name;
+        characterList.connect('character-selected',
+                              Lang.bind(this, this._handleCharacterSelected));
 
-        let scroll = characterList.get_child_by_name('character-list');
-        let widget = scroll.get_child().get_child();
-        widget.connect('character-selected',
-                       Lang.bind(this, this._handleCharacterSelected));
+        this._characterLists[name] = characterList;
+        return characterList;
+    },
+
+    _createRecentCharacterList: function(name, accessible_name, category) {
+        let characterList = new CharacterList.RecentCharacterListView({
+            fontFilter: this._fontFilter,
+            category: category
+        });
+        characterList.get_accessible().accessible_name = accessible_name;
+        characterList.connect('character-selected',
+                              Lang.bind(this, this._handleCharacterSelected));
 
         this._characterLists[name] = characterList;
         return characterList;
@@ -314,21 +423,23 @@ const MainView = new Lang.Class({
     },
 
     setPage: function(category) {
-        let characterList = this.get_child_by_name(category.name);
-        characterList.setFilterFont(this._filterFontFamily);
-
         if (category.name == 'recent') {
             if (this.recentCharacters.length == 0)
-                characterList.visible_child_name = 'empty-recent';
+                this.visible_child_name = 'empty-recent';
             else {
-                characterList.setCharacters(this.recentCharacters);
-                characterList.updateCharacterList();
+                let categories = this._categoryListView.getCategoryList();
+                for (let i in categories) {
+                    let category = categories[i];
+                    let characterList = this._recentCharacterLists[category.name];
+                    characterList.setCharacters(this.recentCharacters);
+                }
+                this.visible_child_name = 'recent';
             }
         } else {
+            let characterList = this.get_child_by_name(category.name);
             characterList.searchByCategory(category);
+            this.visible_child = characterList;
         }
-
-        this.visible_child = characterList;
     },
 
     addToRecent: function(uc) {
@@ -343,17 +454,21 @@ const MainView = new Lang.Class({
         }
     },
 
-    _handleCharacterSelected: function(widget, uc) {
+    _addToRecent: function(widget, uc) {
         this.addToRecent(uc);
+    },
 
+    _handleCharacterSelected: function(widget, uc) {
         let dialog = new Character.CharacterDialog({
             character: uc,
             modal: true,
             transient_for: this.get_toplevel(),
-            fontDescription: this.visible_child.getFontDescription()
+            fontDescription: this._fontFilter.fontDescription
         });
 
         dialog.show();
+        dialog.connect('character-copied',
+                       Lang.bind(this, this._addToRecent));
         dialog.connect('response', function(self, response_id) {
             if (response_id == Gtk.ResponseType.CLOSE)
                 dialog.destroy();
